-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua = { }
npcg.functua = functua

--[[ getOrDefault
function functua.getOrDefault(list, key, default)
  return (list or { })[key] or default
end
]]

--[[ functua.getOrNil
function getOrNil(list, key)
  return (list or { })[key]
end
]]

-- temporary hack
NPCG = npcg
npcg.generators = { }
npcg.data = { }
npcg.tests = { }

-------------------------------------------------------------------------------------
-- | Prelude | ----------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- id :: a -> a
-- Returns the input.
function functua.id(x)
  return x
end

-- replicate :: int -> a -> { a }
-- Returns a list of length n where every element is x.
function functua.replicate(n, x)
	local output = { }
	for i=1, n do table.insert(output, x) end
	return output
end

-- iterate :: (a -> ()) -> int -> int -> ()
-- Performs a certain action incrementally, starting from x, then x+1 until (inclusive) n.
function functua.iterate(action, x, n)
  for i=x, n do
    action(i)
  end
end

-- forEach :: (a -> ()) -> { a } -> ()
-- Performs the same action on every element in a list.
function functua.forEach(a, xs)
	for _,x in ipairs(xs) do
		a(x)
	end
end

-- forEach :: (a -> ()) -> { a } -> ()
-- Performs the same action on every element in a list.
function functua.forEachT(a, xs)
	for _,x in pairs(xs) do
		a(x)
	end
end

--[[ forEachKey
function functua.forEachKey(a, xs)
	for k,_ in pairs(xs) do
		a(k)
	end
end
]]

--[[ forEachKeyValuePair
function functua.forEachKeyValuePair(a, xs)
	for k,v in pairs(xs) do
		a(k,v)
	end
end
]]



-- keys :: { k,_ } -> { k }
-- Returns a list of keys from a table.
function functua.keys(xs)
  local output = { }
  for k,_ in pairs(xs) do
    table.insert(output, k)
  end
  return output
end

-- values :: { _,v } -> { v }
-- Returns a list of values from a table.
function functua.values(xs)
  local output = { }
  for _,v in pairs(xs) do
    table.insert(output, v)
  end
  return output
end

-- insert :: { v } -> v -> { v }
-- Returns a list comprised of a given list and a given value.
function functua.insert(xs, x)
	local output = functua.map(functua.id, xs)
	table.insert(output, x)
	return output
end

-- map :: (a -> b) -> { k, a } -> { k, b }
-- Applies a function to every item in a table and returns the result.
-- Ordering is preserved.
function functua.map(f, xs)
	local output = { }
	for k,v in pairs(xs) do
		output[k] = f(v)
	end
	return output
end

-- map2 :: (k -> v -> v) -> { k, v }
-- Applies a function to every key and value in a table and returns the result.
function functua.map2(f, xs)
	local output = { }
	for k,v in pairs(xs) do
		output[k] = f(k,v)
	end
	return output
end

-- clone :: { k,v } -> { k,v }
-- Returns a clone of a table.
function functua.clone(xs)
	return functua.map(functua.id, xs)
end
--[[ mapKeys
function functua.mapKeys(f, xs)
  local output = { }
  for k,_ in pairs(xs) do
    table.insert(output, f(k))
  end
  return output
end
]]

--[[ mapKeyValuePairs
function functua.mapKeyValuePairs(f, xs)
  local output = { }
  for k,v in pairs(xs) do
    table.insert(output, f(KeyValuePair(k,v)))
  end
  return output
end
]]

-- filter :: (a -> bool) -> { _,a } -> { a }
-- Returns a list of all elements in a list that satisfy a predicate.
function functua.filter(p, xs)
	local output = { }
	for _,v in pairs(xs) do
		if p(v) then 
			table.insert(output, v)
		end
	end
	return output
end

-- filter2 :: (a -> bool) -> { a,b } -> { b }
-- Returns list of key-value-pairs of which the key satisfies a predicate.
function functua.filter2(p, xs)
  local output = { }
	for k,v in pairs(xs) do
		if p(k) then 
			output[k] = v
		end
	end
	return output
end

-- filter3 :: (b -> bool) -> { a,b } -> { a,b }
-- Returns a list of all key-value-pairs of which the value satisfies a predicate.
function functua.filter3(p, xs)
  local output = { }
	for k,v in pairs(xs) do
		if p(v) then 
			output[k] = v
		end
	end
	return output
end

-- filterKeysByValue :: (a -> bool) -> { k -> a } -> { k }
-- Returns a list of all keys in a list where the value satisfies a certain predicate.
function functua.filterKeysByValue(p, xs)
  local output = { }
  for k,v in pairs(xs) do
    if p(v) then 
      table.insert(output, k)
    end
  end
  return output
end

-- foldr :: (a -> b -> b) -> b -> { a } -> b
-- Applies an accumulator function on every element of a given list.
function functua.foldr(op, val, xs)
	for _,v in ipairs(xs) do
		val = op(val, v)
	end
	return val
end

-- foldr2 :: (a -> b -> b) -> b -> { _,a } -> b
-- Applies an accumulator function on every element of a given table.
function functua.foldr2(op, val, xs)
	for _,v in pairs(xs) do
		val = op(val, v)
	end
	return val
end

--[[ foldrByKey
function functua.foldrByKey(op, val, xs)
  for k,_ in pairs(xs) do
    val = op(val, k)
  end
  return val
end
]]

--[[ foldrByKeyValuePair
function functua.foldrByKeyValuePair(op, val, xs)
  for k,v in pairs(xs) do
    val = op(val, KeyValuePair(k,v))
  end
  return val
end
]]

-- concat :: { _,{ _,b } } -> { _,b }
-- Concatenates a list of lists.
function functua.concat(xs)
	local output = { }
	for _,v in pairs(xs) do
		for _,vv in pairs(v) do
			table.insert(output, vv)
		end
	end
	return output
end

-- concat :: { _,{ a,_ } } -> { a,_ }
-- Concatenates a list of lists.
function functua.concatKeys(xs)
	local output = { }
	for _,v in pairs(xs) do
		for kv,_ in pairs(v) do
			output[kv] = 1
		end
	end
	return functua.keys(output)
end

-- any :: (a -> bool) -> { a } -> bool
-- Returns whether any of the elements in a list satisfies a predicate.
function functua.any(p, xs)
	for _,v in pairs(xs) do
		if p(v) then
			return true
		end
	end
	return false
end


function functua.all(p, xs)
	return not functua.any(
		function(x) return not p(x) end, 
		xs)
end


--[[ none
function functua.none(p, xs)
	return not functua.any(p, xs)
end
]]

-- contains :: a -> { a } -> bool
-- Returns whether a value is contained in a list.
function functua.contains(val, xs)
	return functua.any(
		function(x) return x == val end, 
		xs)
end

-- containsAny :: { _,b } --> { b } --> bool
-- Returns whether a list contains any of the values of a second list.
function functua.containsAny(xs, values)
  return functua.any(
    function(x) return functua.contains(x, xs) end,
    values)
end

function functua.containsAll(xs, values)
  return functua.all(
    function(x) return functua.contains(x, xs) end,
    values)
end

--[[
function functua.containsKey(key, xs)
	for k,_ in pairs(xs) do
		if k == key then return true end
	end
	return false
end
]]

-- const :: a -> (b -> a)
-- Returns a function that always returns x.
function functua.const(x)
  return function(...) return x end
end

-- invokeAll :: { (() -> ()) } -> ()
-- Invokes all actions in a list.
function functua.invokeAll(xs)
  functua.forEach(function(x) x() end, xs)
end

-- count :: { a } -> int
-- Counts the number of elements in a list.
function functua.count(list)
	return functua.foldr(
    function(accum, x) return accum + 1 end, 0, 
    list
  )
end

-- sum :: Num a => { a } -> a
-- Computes the sum of the items in the list.
-- Items must be numbers.
function functua.sum(list)
  return functua.foldr(
    function(accum, x) return accum + x end,
    0,
    list
  )
end

-- sumBy :: Num b => { a } -> (a -> b) -> b
-- Computes the sum of the items in the list, using a function to convert the items into numbers.
function functua.sumBy(list, by)
  return functua.foldr2(
    function(accum, x) return accum + by(x) end,
    0,
    list
  )
end

-- mergeTables :: { a } -> { a } -> { a }
--[[ Merges an arbitrary amount of arrays into one.
-- Ensures:
-- 1) ({1},{2}) = {1,2}
-- 2) ({ },{2}) = {  2}
-- 3) ({1},{ }) = {  1}
-- 4) (nil,{2}) = {  2}
-- 5) ({1},nil) = {  1} ]]
function functua.mergeTables(...)
	local output = { }
	for _,t in pairs({...}) do  
		for _,v in pairs(t or { }) do -- <-- only takes values, indicating it's a list, but also uses pairs, which indicates it should be a table..
			table.insert(output, v)
		end
	end
	return output
end

function functua.mergeTables2(...)
  local output = { }
  for _,t in ipairs({...}) do
    for k,v in pairs(t or { }) do
      output[k] = v
    end
  end
  return output
end

function functua.mergeTablesWithoutDuplicates(...)
  local output = { }
	for _,t in pairs({...}) do  -- <-- only takes values, indicating it's a list, but also uses pairs, which indicates it should be a table..
		for _,v in pairs(t or { }) do
			if not functua.contains(v, output) then
        table.insert(output, v)
      end
		end
	end
	return output
end

function functua.mergeTablesWithoutDuplicates2(...)
  local output = { }
	for _,t in pairs({...}) do  -- <-- only takes values, indicating it's a list, but also uses pairs, which indicates it should be a table..
		for k,v in pairs(t or { }) do
			--if not contains(k, output) then
        output[k] = v
      --end
		end
	end
	return output
end

-- intersect :: { a } -> { a } -> { a } 
-- Returns the intersection between two lists, aka the elements that are both in xs and ys.
-- intersect( {1},   {1,1} ) returns {1}
-- intersect( {1,1}, {1}   ) returns {1,1}
function functua.intersect(xs, ys)
  -- Rough way to calculate an intersection between two lists.
  local set = { }
  local output = { }
  for _,y in ipairs(ys) do
    set[y] = true
  end
  for _,x in ipairs(xs) do
    if set[x] then
      table.insert(output, x)
    end
  end
  return output
end

-- without :: { a } -> { a } -> { a }
-- Removes all occurances of certain values from a list.
function functua.without(list, values)
  return functua.filter(
    function(x) return not functua.contains(x, values) end,
    list)
end

--[[function functua.find(p, xs)
  forEach(
    function(x)
      if p(x)then
        return x
      end
    end,
    xs)
  return nil
end]]

-- findKey :: 
-- Returns
function functua.findKey(p, xs)
  for k,v in pairs(xs) do
    if p(v) then return k end
  end
  return nil
end

-- skipWhile :: (a -> bool) -> { a } -> { a }
-- Returns a list where all items until the predicate is satisfied are ignored. The element where the predicate is true is included.
function functua.skipWhile(p, xs)
  local output = { }
  local found = false
  for _,v in ipairs(xs) do
    if p(v) then
      found = true
    end
    if found then 
      table.insert(output, v)
    end
  end
  return output
end

-- takeWhile :: (a -> bool) -> { a } -> { a }
-- Returns a list where all items after the predicate is satisfied are ignored. The element where the predicate is true is not included.
function functua.takeWhile(p, xs)
  local output = { }
  local found = false
  for _,v in ipairs(xs) do
    if   p(v) then return output
    else           table.insert(output, v) 
    end
  end
  return output
end

--[[ flip
function functua.flip(f)
  return function(x,y) return f(y,x) end
end
]]

-- shift :: { k, v } -> { { key :: k, value :: v } }
-- Returns a list of key-value-pairs from a table of key-value-pairs.
-- Note: The original k is now in x.key and the original v is now in x.value.
function functua.shift(xs)
	local output = { }
	
	-- No existing map function suffices.
	for k,v in pairs(xs) do
		table.insert(output, { key=k, value=v })
	end
	
	return output
end

-- unshift :: { { key :: k, value :: v } } -> { k, v }
-- Returns a table list of key-value-pairs.
function functua.unshift(xs)
	local output = { }
	
	-- No existing map function suffices.
	for _,v in ipairs(xs) do
		output[v.key] = v.value
	end
	
	return output
end

function functua.equals(x, y)
  return x == y
end

function functua.composeEqualsFunction(y)
  return function(x) return x == y end
end

-- End of File.